"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os

from typing import Tuple
from typing import List
from typing import Dict
from typing import Callable
from typing import Union
from typing import Type
from typing import Any

from ...tfile import TorrentEntryAttrs
from ...tfile import Torrent

from .. import BasePlugin
from .. import get_classes


# =====
class NoSuchTorrentError(Exception):
    pass


# =====
def hash_or_torrent(method: Callable) -> Callable:
    def wrap(self: "BaseClient", torrent: Union[Torrent, str], *args: Any, **kwargs: Any) -> Any:
        torrent_hash = (torrent.get_hash() if isinstance(torrent, Torrent) else torrent)
        return method(self, torrent_hash, *args, **kwargs)
    return wrap


def check_torrent_accessible(method: Callable) -> Callable:
    def wrap(self: "BaseClient", torrent: Torrent, prefix: str="") -> Any:
        path = torrent.get_path()
        assert path is not None, "Required Torrent() with local file"
        open(path, "rb").close()  # Check accessible file
        if prefix:
            os.listdir(prefix)  # Check accessible prefix
        return method(self, torrent, prefix)
    return wrap


def build_files(prefix: str, flist: List[Tuple[str, int]]) -> Dict[str, TorrentEntryAttrs]:
    files: Dict[str, TorrentEntryAttrs] = {}
    for (path, size) in flist:
        path_list = path.split(os.path.sep)
        name = None
        for index in range(len(path_list)):
            name = os.path.join(prefix, os.path.sep.join(path_list[0:index + 1]))
            files[name] = TorrentEntryAttrs.new_dir()
        assert name is not None
        files[name] = TorrentEntryAttrs.new_file(size)
    return files


class BaseClient(BasePlugin):
    def __init__(self, **_: Any) -> None:  # pylint: disable=super-init-not-called
        pass

    @hash_or_torrent
    def remove_torrent(self, torrent_hash: str) -> None:
        raise NotImplementedError

    @check_torrent_accessible
    def load_torrent(self, torrent: Torrent, prefix: str="") -> None:
        raise NotImplementedError

    def get_hashes(self) -> List[str]:
        raise NotImplementedError

    @hash_or_torrent
    def has_torrent(self, torrent_hash: str) -> bool:
        raise NotImplementedError

    @hash_or_torrent
    def get_torrent_path(self, torrent_hash: str) -> str:
        raise NotImplementedError

    @hash_or_torrent
    def get_data_prefix(self, torrent_hash: str) -> str:
        raise NotImplementedError

    def get_data_prefix_default(self) -> str:
        raise NotImplementedError

    # ===

    @hash_or_torrent
    def get_full_path(self, torrent_hash: str) -> str:
        raise NotImplementedError

    @hash_or_torrent
    def get_file_name(self, torrent_hash: str) -> str:
        raise NotImplementedError

    @hash_or_torrent
    def is_single_file(self, torrent_hash: str) -> bool:
        raise NotImplementedError

    @hash_or_torrent
    def get_files(self, torrent_hash: str, on_fs: bool=False) -> Dict[str, TorrentEntryAttrs]:
        raise NotImplementedError


class WithCustoms(BaseClient):
    def __init__(self, **_: Any) -> None:  # pylint: disable=super-init-not-called
        pass

    @classmethod
    def get_custom_keys(cls) -> List[str]:
        raise NotImplementedError

    @hash_or_torrent
    def set_customs(self, torrent_hash: str, customs: Dict[str, str]) -> None:
        raise NotImplementedError

    @hash_or_torrent
    def get_customs(self, torrent_hash: str, keys: List[str]) -> Dict[str, str]:
        raise NotImplementedError


# =====
def get_client_class(name: str) -> Type[BaseClient]:
    return get_classes("clients")[name]  # type: ignore



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import re
import socket
import urllib.request
import urllib.parse
import urllib.error
import http.client
import http.cookiejar
import json
import datetime

from typing import Dict
from typing import Pattern
from typing import Match
from typing import Callable
from typing import NamedTuple
from typing import Optional
from typing import Type
from typing import Any

import pytz

from ...optconf import Option
from ...optconf import SecretOption

from ...tfile import Torrent
from ...tfile import is_valid_torrent_data
from ...tfile import decode_torrent_data

from ... import web

from .. import BasePlugin
from .. import get_classes


# =====
class TrackerError(Exception):
    pass


class AuthError(TrackerError):
    pass


class LogicError(TrackerError):
    pass


class NetworkError(TrackerError):
    def __init__(self, sub: BaseException) -> None:
        super().__init__()
        self._sub = sub

    def __str__(self) -> str:
        return "{}: {}".format(type(self._sub).__name__, str(self._sub))


# =====
def _assert(exception: Type[TrackerError], arg: Any, msg: str="") -> None:
    if not arg:
        raise exception(msg)


class BaseTracker(BasePlugin):  # pylint: disable=too-many-instance-attributes
    _SITE_VERSION = 0
    _SITE_ENCODING = "utf-8"
    _SITE_RETRY_CODES = [500, 502, 503]

    _SITE_FINGERPRINT_URL = __D_SITE_FINGERPRINT_URL = ""
    _SITE_FINGERPRINT_TEXT = __D_SITE_FINGERPRINT_TEXT = ""

    _COMMENT_REGEXP = __D_COMMENT_REGEXP = re.compile(r"(?P<torrent_id>.*)")

    def __init__(  # pylint: disable=super-init-not-called
        self,
        timeout: float,
        retries: int,
        retries_sleep: float,
        user_agent: str,
        proxy_url: str,
        check_version: bool,
        check_fingerprint: bool,
        **_: Any,
    ) -> None:

        assert self._SITE_FINGERPRINT_URL != self.__D_SITE_FINGERPRINT_URL
        assert self._SITE_FINGERPRINT_TEXT != self.__D_SITE_FINGERPRINT_TEXT

        assert self._COMMENT_REGEXP.pattern != self.__D_COMMENT_REGEXP.pattern

        self.__timeout = timeout
        self.__retries = retries
        self.__retries_sleep = retries_sleep
        self.__user_agent = user_agent
        self.__proxy_url = proxy_url
        self.__check_version = check_version
        self.__check_fingerprint = check_fingerprint

        self.__opener: Optional[urllib.request.OpenerDirector] = None

        self._cookie_jar: Optional[http.cookiejar.CookieJar] = None

    @classmethod
    def get_options(cls) -> Dict[str, Option]:
        return {
            "timeout":           Option(default=10.0, help="Timeout for HTTP client"),
            "retries":           Option(default=20, help="The number of retries to handle tracker-specific HTTP errors"),
            "retries_sleep":     Option(default=1.0, help="Sleep interval between failed retries"),
            "user_agent":        Option(default="Mozilla/5.0", help="User-Agent for site"),
            "proxy_url":         Option(default="", help="URL of HTTP/SOCKS4/SOCKS5 proxy"),
            "check_fingerprint": Option(default=True, help="Check the site fingerprint"),
            "check_version":     Option(default=True, help="Check the tracker version from GitHub"),
        }

    def test(self) -> None:
        if self.__check_fingerprint or self.__check_version:
            opener = web.build_opener(self.__proxy_url)
            info = self.__get_upstream_info(opener)
        if self.__check_fingerprint:
            self.__test_fingerprint(info["fingerprint"], opener)
        if self.__check_version:
            self.__test_version(info["version"])

    def is_matched_for(self, torrent: Torrent) -> bool:
        return (self._COMMENT_REGEXP.match(torrent.get_comment()) is not None)

    def fetch_new_data(self, torrent: Torrent) -> bytes:
        raise NotImplementedError

    # ===

    def _encode(self, arg: str) -> bytes:
        return arg.encode(self._SITE_ENCODING)

    def _decode(self, arg: bytes) -> str:
        return arg.decode(self._SITE_ENCODING)

    # ===

    def _init_opener(self, with_cookies: bool) -> None:
        assert not self.__opener
        if with_cookies:
            self._cookie_jar = http.cookiejar.CookieJar()
            self.__opener = web.build_opener(self.__proxy_url, self._cookie_jar)
        else:
            self.__opener = web.build_opener(self.__proxy_url)

    def _read_url(self, *args: Any, **kwargs: Any) -> bytes:
        if not kwargs.get("opener"):
            kwargs["opener"] = self.__opener
        try:
            return self.__read_url_nofe(*args, **kwargs)
        except (
            socket.timeout,
            urllib.error.HTTPError,
            urllib.error.URLError,
            http.client.IncompleteRead,
            http.client.BadStatusLine,
            ConnectionResetError,
        ) as err:
            raise NetworkError(err)

    def __read_url_nofe(
        self,
        url: str,
        *,
        data: Optional[bytes]=None,
        headers: Optional[Dict[str, str]]=None,
        opener: urllib.request.OpenerDirector=None,
    ) -> bytes:

        assert opener
        headers = (headers or {})
        headers.setdefault("User-Agent", self.__user_agent)
        return web.read_url(
            opener=opener,
            url=url,
            data=data,
            headers=headers,
            timeout=self.__timeout,
            retries=self.__retries,
            retries_sleep=self.__retries_sleep,
            retry_codes=self._SITE_RETRY_CODES,
        )

    # ===

    def _assert_logic(self, arg: Any, *args: Any) -> None:
        _assert(LogicError, arg, *args)

    def _assert_logic_re_match(self, regexp: Pattern[str], text: str, msg: str) -> Match[str]:
        match = regexp.match(text)
        self._assert_logic(bool(match), msg)
        return match  # type: ignore

    def _assert_logic_re_search(self, regexp: Pattern[str], text: str, msg: str) -> Match[str]:
        match = regexp.search(text)
        self._assert_logic(bool(match), msg)
        return match  # type: ignore

    def _assert_match(self, torrent: Torrent) -> str:
        return self._assert_logic_re_match(
            regexp=self._COMMENT_REGEXP,
            text=torrent.get_comment(),
            msg="No match with torrent's comment",
        ).group("torrent_id")

    def _assert_valid_data(self, data: bytes, target: str="torrent") -> bytes:
        msg = "Received an invalid {} data: {} ...".format(target, repr(data[:20]))
        self._assert_logic(is_valid_torrent_data(data), msg)
        return data

    # ===

    def __get_upstream_info(self, opener: urllib.request.OpenerDirector) -> Dict:
        try:
            return json.loads(self.__read_url_nofe(
                url="https://raw.githubusercontent.com/mdevaev/emonoda/master/trackers/{}.json".format(self.PLUGIN_NAME),
                opener=opener,
            ).decode("utf-8"))
        except urllib.error.HTTPError as err:
            if err.code == 404:
                return self._get_local_info()
            raise

    @classmethod
    def _get_local_info(cls) -> Dict:  # Public for Makefile
        return {
            "version": cls._SITE_VERSION,
            "fingerprint": {
                "url":      cls._SITE_FINGERPRINT_URL,
                "encoding": cls._SITE_ENCODING,
                "text":     cls._SITE_FINGERPRINT_TEXT,
            },
        }

    def __test_fingerprint(self, fingerprint: Dict[str, str], opener: urllib.request.OpenerDirector) -> None:
        data = self._read_url(fingerprint["url"], opener=opener)
        msg = "Invalid site body, maybe tracker is blocked"
        try:
            page = data.decode(fingerprint["encoding"])
        except UnicodeDecodeError:
            raise TrackerError(msg)
        _assert(TrackerError, fingerprint["text"] in page, msg)

    def __test_version(self, upstream: int) -> None:
        _assert(
            TrackerError,
            self._SITE_VERSION >= upstream,
            "Tracker plugin is outdated (ver. local:{}, upstream:{}). I recommend to update the program".format(
                self._SITE_VERSION,
                upstream,
            ),
        )


class WithLogin(BaseTracker):
    def __init__(self, user: str, passwd: str, **_: Any) -> None:  # pylint: disable=super-init-not-called
        self._user = user
        self._passwd = passwd

    @classmethod
    def get_options(cls) -> Dict[str, Option]:
        return {
            "user":   Option(default="", help="Site login"),
            "passwd": SecretOption(default="", help="Site password"),
        }

    def login(self) -> None:
        raise NotImplementedError

    def _login_using_post(self, url: str, post: Dict[str, bytes], ok_text: str) -> None:
        self._assert_required_user_passwd()
        page = self._decode(self._read_url(url, data=self._encode(urllib.parse.urlencode(post))))
        self._assert_auth(ok_text in page, "Invalid user or password")

    def _assert_auth(self, *args: Any) -> None:
        _assert(AuthError, *args)  # pylint: disable=no-value-for-parameter

    def _assert_required_user_passwd(self) -> None:
        self._assert_auth(bool(self._user), "Required user for site")
        self._assert_auth(bool(self._passwd), "Required password for site")

    def _assert_auth_re_search(self, regexp: Pattern[str], text: str, msg: str) -> Match[str]:
        match = regexp.search(text)
        self._assert_auth(bool(match), msg)
        return match  # type: ignore


class WithCaptcha(BaseTracker):  # pylint: disable=abstract-method
    def __init__(self, captcha_decoder: Callable[[str], str], **_: Any) -> None:  # pylint: disable=super-init-not-called
        self._captcha_decoder = captcha_decoder


# =====
class WithCheckHash(BaseTracker):  # pylint: disable=abstract-method
    _TORRENT_HASH_URL = __D_TORRENT_HASH_URL = "{torrent_id}"
    _TORRENT_HASH_REGEXP = __D_TORRENT_HASH_REGEXP = re.compile(r"(?P<torrent_hash>.*)")

    def __init__(self, **_: Any) -> None:  # pylint: disable=super-init-not-called
        assert self._TORRENT_HASH_URL != self.__D_TORRENT_HASH_URL
        assert self._TORRENT_HASH_REGEXP.pattern != self.__D_TORRENT_HASH_REGEXP.pattern

    def fetch_hash(self, torrent: Torrent) -> str:
        torrent_id = self._assert_match(torrent)
        return self._assert_logic_re_search(
            regexp=self._TORRENT_HASH_REGEXP,
            text=self._decode(self._read_url(self._TORRENT_HASH_URL.format(torrent_id=torrent_id))),
            msg="Hash not found",
        ).group("torrent_hash").strip().lower()


class WithCheckScrape(BaseTracker):  # pylint: disable=abstract-method
    _TORRENT_SCRAPE_URL = __D_TORRENT_SCRAPE_URL = "{scrape_hash}"

    def __init__(self, client_agent: str, **_: Any) -> None:  # pylint: disable=super-init-not-called
        assert self._TORRENT_SCRAPE_URL != self.__D_TORRENT_SCRAPE_URL

        self._client_agent = client_agent

    @classmethod
    def get_options(cls) -> Dict[str, Option]:
        return {
            "client_agent": Option(default="rtorrent/0.9.2/0.13.2", help="User-Agent for tracker"),
        }

    def is_registered(self, torrent: Torrent) -> bool:
        # https://wiki.theory.org/BitTorrentSpecification#Tracker_.27scrape.27_Convention
        self._assert_match(torrent)
        data = self._assert_valid_data(self._read_url(
            url=self._TORRENT_SCRAPE_URL.format(scrape_hash=torrent.get_scrape_hash()),
            headers={"User-Agent": self._client_agent},
        ), target="scrape")
        return (len(decode_torrent_data(data).get("files", {})) != 0)


class WithCheckTime(BaseTracker):
    _TIMEZONE_URL = __D_TIMEZONE_URL = ""
    _TIMEZONE_REGEXP = __D_TIMEZONE_REGEXP = re.compile(r"(?P<torrent_hash>.*)")
    _TIMEZONE_PREFIX = __D_TIMEZONE_PREFIX = ""

    _TIMEZONE_STATIC = __D_TIMEZONE_STATIC = ""

    def __init__(self, timezone: str, **_: Any) -> None:  # pylint: disable=super-init-not-called
        assert (
            self._TIMEZONE_URL != self.__D_TIMEZONE_URL
            and self._TIMEZONE_REGEXP.pattern != self.__D_TIMEZONE_REGEXP.pattern
            and self._TIMEZONE_PREFIX != self.__D_TIMEZONE_PREFIX
        ) or self._TIMEZONE_STATIC != self.__D_TIMEZONE_STATIC

        self._default_timezone = timezone
        self._tzinfo: Optional[datetime.tzinfo] = None

    @classmethod
    def get_options(cls) -> Dict[str, Option]:
        return {
            "timezone": Option(
                default="",
                help="Site timezone, is automatically detected if possible (or manually, 'Europe/Moscow' for example)",
            ),
        }

    def init_tzinfo(self) -> None:
        timezone: Optional[str] = None
        if self._TIMEZONE_STATIC:
            timezone = self._TIMEZONE_STATIC
        else:
            page = self._decode(self._read_url(self._TIMEZONE_URL))
            timezone_match = self._TIMEZONE_REGEXP.search(page)
            if timezone_match is not None:
                timezone = self._TIMEZONE_PREFIX + timezone_match.group("timezone").replace(" ", "")
        self._tzinfo = self._select_tzinfo(timezone)

    def fetch_time(self, torrent: Torrent) -> int:
        raise NotImplementedError

    def _select_tzinfo(self, site_timezone: Optional[str]) -> datetime.tzinfo:
        if not site_timezone or self._default_timezone:
            return self.__get_default_tzinfo()
        try:
            return pytz.timezone(site_timezone)
        except pytz.UnknownTimeZoneError:  # type: ignore
            return self.__get_default_tzinfo()

    def __get_default_tzinfo(self) -> datetime.tzinfo:
        msg = "Can't determine timezone of site, your must configure it manually"
        self._assert_logic(bool(self._default_timezone), msg)
        return pytz.timezone(self._default_timezone)


# =====
class WithFetchByTorrentId(BaseTracker):
    _DOWNLOAD_URL = __D_DOWNLOAD_URL = "{torrent_id}"
    _DOWNLOAD_PAYLOAD: Optional[bytes] = None

    def __init__(self, **_: Any) -> None:  # pylint: disable=super-init-not-called
        assert self._DOWNLOAD_URL != self.__D_DOWNLOAD_URL

    def fetch_new_data(self, torrent: Torrent) -> bytes:
        torrent_id = self._assert_match(torrent)
        return self._assert_valid_data(self._read_url(
            url=self._DOWNLOAD_URL.format(torrent_id=torrent_id),
            data=self._DOWNLOAD_PAYLOAD,
        ))


class WithFetchByDownloadId(BaseTracker):
    _DOWNLOAD_ID_URL = __D_DOWNLOAD_ID_URL = "{torrent_id}"
    _DOWNLOAD_ID_REGEXP = __D_DOWNLOAD_ID_REGEXP = re.compile(r"(?P<download_id>.*)")
    _DOWNLOAD_URL = __D_DOWNLOAD_URL = "{download_id}"
    _DOWNLOAD_PAYLOAD: Optional[bytes] = None

    def __init__(self, **_: Any) -> None:  # pylint: disable=super-init-not-called
        assert self._DOWNLOAD_ID_URL != self.__D_DOWNLOAD_ID_URL
        assert self._DOWNLOAD_ID_REGEXP.pattern != self.__D_DOWNLOAD_ID_REGEXP.pattern
        assert self._DOWNLOAD_URL != self.__D_DOWNLOAD_URL

    def fetch_new_data(self, torrent: Torrent) -> bytes:
        torrent_id = self._assert_match(torrent)

        dl_id = self._assert_logic_re_search(
            regexp=self._DOWNLOAD_ID_REGEXP,
            text=self._decode(self._read_url(self._DOWNLOAD_ID_URL.format(torrent_id=torrent_id))),
            msg="Unknown download_id",
        ).group("download_id")

        return self._assert_valid_data(self._read_url(
            url=self._DOWNLOAD_URL.format(download_id=dl_id),
            data=self._DOWNLOAD_PAYLOAD,
        ))


class TrackerStat(NamedTuple):
    seeders: int = 0
    leechers: int = 0


class WithStat(BaseTracker):  # pylint: disable=abstract-method
    _STAT_URL = __D_STAT_URL = "{torrent_id}"
    _STAT_SEEDERS_REGEXP = __D_STAT_SEEDERS_REGEXP = re.compile(r"(?P<seeders>.*)")
    _STAT_LEECHERS_REGEXP = __D_STAT_LEECHERS_REGEXP = re.compile(r"(?P<leechers>.*)")

    def __init__(self, **_: Any) -> None:  # pylint: disable=super-init-not-called
        assert self._STAT_URL != self.__D_STAT_URL
        assert self._STAT_SEEDERS_REGEXP.pattern != self.__D_STAT_SEEDERS_REGEXP.pattern
        assert self._STAT_LEECHERS_REGEXP.pattern != self.__D_STAT_LEECHERS_REGEXP.pattern

    def fetch_stat(self, torrent: Torrent) -> TrackerStat:
        torrent_id = self._assert_match(torrent)
        page = self._decode(self._read_url(self._STAT_URL.format(torrent_id=torrent_id)))
        return TrackerStat(
            seeders=self.__parse_stat_int(page, self._STAT_SEEDERS_REGEXP, "seeders"),
            leechers=self.__parse_stat_int(page, self._STAT_LEECHERS_REGEXP, "leechers"),
        )

    def __parse_stat_int(self, page: str, regexp: Pattern, group: str) -> int:
        match = regexp.search(page)
        if match:
            return int(match.group(group))
        return 0


# =====
def get_tracker_class(name: str) -> Type[BaseTracker]:
    return get_classes("trackers")[name]  # type: ignore



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os
import base64

from typing import List
from typing import Dict
from typing import Optional
from typing import Union
from typing import Any

from ...optconf import Option
from ...optconf import SecretOption

from ...tfile import TorrentEntryAttrs
from ...tfile import Torrent

from . import BaseClient
from . import NoSuchTorrentError
from . import hash_or_torrent
from . import check_torrent_accessible
from . import build_files

try:
    import transmissionrpc
except ImportError:
    transmissionrpc = None


# =====
class Plugin(BaseClient):
    # API description:
    #   http://pythonhosted.org/transmissionrpc/
    #   https://trac.transmissionbt.com/browser/trunk/extras/rpc-spec.txt

    PLUGIN_NAME = "transmission"

    def __init__(  # pylint: disable=super-init-not-called
        self,
        url: str,
        user: str,
        passwd: str,
        timeout: float,
        **kwargs: Any,
    ) -> None:

        self._init_bases(**kwargs)

        if transmissionrpc is None:
            raise RuntimeError("Required module transmissionrpc")

        self.__client = transmissionrpc.Client(
            address=url,
            user=(user or None),
            password=(passwd or None),
            timeout=timeout,
        )

    @classmethod
    def get_options(cls) -> Dict[str, Option]:
        return cls._get_merged_options({
            "url":     Option(default="http://localhost:9091/transmission/rpc", help="Transmission HTTP-RPC URL"),
            "timeout": Option(default=10.0, help="Timeout for HTTP-RPC"),
            "user":    Option(default="", help="HTTP login"),
            "passwd":  SecretOption(default="", help="HTTP password"),
        })

    # ===

    @hash_or_torrent
    def remove_torrent(self, torrent_hash: str) -> None:
        self.__get_torrent_obj(torrent_hash)  # XXX: raise NoSuchTorrentError if torrent does not exist
        self.__client.remove_torrent(torrent_hash)

    @check_torrent_accessible
    def load_torrent(self, torrent: Torrent, prefix: str="") -> None:
        kwargs: Dict[str, Union[bool, str]] = {"paused": False}
        if prefix:
            kwargs["download_dir"] = prefix
        self.__client.add_torrent(base64.b64encode(torrent.get_data()).decode("utf-8"), **kwargs)

    @hash_or_torrent
    def has_torrent(self, torrent_hash: str) -> bool:
        try:
            self.__get_torrent_obj(torrent_hash)
            return True
        except NoSuchTorrentError:
            return False

    def get_hashes(self) -> List[str]:
        return [
            str(item.hashString.lower())
            for item in self.__client.get_torrents(arguments=("id", "hashString"))
        ]

    @hash_or_torrent
    def get_torrent_path(self, torrent_hash: str) -> str:
        return self.__get_torrent_prop(torrent_hash, "torrentFile")

    @hash_or_torrent
    def get_data_prefix(self, torrent_hash: str) -> str:
        return self.__get_torrent_prop(torrent_hash, "downloadDir")

    def get_data_prefix_default(self) -> str:
        session = self.__client.get_session()
        assert session is not None
        return session.download_dir

    # ===

    @hash_or_torrent
    def get_full_path(self, torrent_hash: str) -> str:
        torrent_obj = self.__get_torrent_obj(torrent_hash, ["name", "downloadDir"])
        return os.path.join(torrent_obj.downloadDir, torrent_obj.name)

    @hash_or_torrent
    def get_file_name(self, torrent_hash: str) -> str:
        return self.__get_torrent_prop(torrent_hash, "name")

    @hash_or_torrent
    def is_single_file(self, torrent_hash: str) -> bool:
        files = self.__get_files(torrent_hash)
        if len(files) > 1:
            return False
        return (os.path.sep not in list(files.values())[0]["name"])

    @hash_or_torrent
    def get_files(self, torrent_hash: str, on_fs: bool=False) -> Dict[str, TorrentEntryAttrs]:
        prefix = (self.get_data_prefix(torrent_hash) if on_fs else "")
        flist = [
            (item["name"], item["size"])
            for item in self.__get_files(torrent_hash).values()
        ]
        return build_files(prefix, flist)

    # ===

    def __get_torrent_prop(self, torrent_hash: str, prop: str) -> Any:
        return getattr(self.__get_torrent_obj(torrent_hash, [prop]), prop)

    def __get_torrent_obj(self, torrent_hash: str, props: Optional[List[str]]=None) -> Any:
        props = list(set(props or []).union(["id", "hashString"]))
        try:
            torrent_obj = self.__client.get_torrent(torrent_hash, arguments=props)
        except KeyError as err:
            if str(err) == "\'Torrent not found in result\'":
                raise NoSuchTorrentError("Unknown torrent hash")
            raise
        assert str(torrent_obj.hashString).lower() == torrent_hash
        return torrent_obj

    def __get_files(self, torrent_hash: str) -> Dict:
        files = self.__client.get_files(torrent_hash)
        if len(files) == 0:
            raise NoSuchTorrentError("Unknown torrent hash")
        assert len(files) == 1
        files = list(files.values())[0]
        assert len(files) > 0
        return files



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import os
import time
import xmlrpc.client

from typing import List
from typing import Dict
from typing import Callable
from typing import Any

from ...optconf import Option

from ...tfile import TorrentEntryAttrs
from ...tfile import Torrent

from . import WithCustoms
from . import NoSuchTorrentError
from . import hash_or_torrent
from . import check_torrent_accessible
from . import build_files


# =====
_XMLRPC_UNKNOWN_HASH = -501


def _catch_unknown_torrent(method: Callable) -> Callable:
    def wrap(self: WithCustoms, *args: Any, **kwargs: Any) -> Any:
        try:
            return method(self, *args, **kwargs)
        except xmlrpc.client.Fault as err:
            if err.faultCode == _XMLRPC_UNKNOWN_HASH:
                raise NoSuchTorrentError("Unknown torrent hash")
            raise
    return wrap


# =====
class Plugin(WithCustoms):
    # API description: http://code.google.com/p/gi-torrent/wiki/rTorrent_XMLRPC_reference

    PLUGIN_NAME = "rtorrent"

    def __init__(  # pylint:disable=super-init-not-called
        self,
        url: str,
        load_retries: int,
        retries_sleep: float,
        xmlrpc_size_limit: int,
        **kwargs: Any,
    ) -> None:

        self._init_bases(**kwargs)

        self.__load_retries = load_retries
        self.__retries_sleep = retries_sleep
        self.__server = xmlrpc.client.ServerProxy(url)
        self.__server.set_xmlrpc_size_limit(xmlrpc_size_limit)

    @classmethod
    def get_options(cls) -> Dict[str, Option]:
        return cls._get_merged_options({
            "url":               Option(default="http://localhost/RPC2", help="XMLRPC mountpoint"),
            "load_retries":      Option(default=10, help="The number of retries to load the torrent"),
            "retries_sleep":     Option(default=1.0, help="Sleep interval between the load retries"),
            "xmlrpc_size_limit": Option(default=67108863, help="Max XMLRPC data size"),
        })

    # ===

    @hash_or_torrent
    @_catch_unknown_torrent
    def remove_torrent(self, torrent_hash: str) -> None:
        self.__server.d.erase(torrent_hash)

    @check_torrent_accessible
    def load_torrent(self, torrent: Torrent, prefix: str="") -> None:
        torrent_hash = torrent.get_hash()
        # XXX: https://github.com/rakshasa/rtorrent/issues/22
        # All load_* calls re asynchronous, so we need to wait until the load of torrent files is complete.
        self.__server.load_raw(torrent.get_data())
        retries = self.__load_retries
        while True:
            try:
                assert self.__server.d.get_hash(torrent_hash).lower() == torrent_hash
                break
            except xmlrpc.client.Fault as err:
                if err.faultCode != _XMLRPC_UNKNOWN_HASH:
                    raise
                if retries == 0:
                    raise RuntimeError("Timed out torrent uploads after {} seconds".format(
                        self.__load_retries * self.__retries_sleep))
                retries -= 1
                time.sleep(self.__retries_sleep)

        if prefix:
            self.__server.d.set_directory(torrent_hash, prefix)
        self.__server.d.start(torrent_hash)

    @hash_or_torrent
    def has_torrent(self, torrent_hash: str) -> bool:
        try:
            assert self.__server.d.get_hash(torrent_hash).lower() == torrent_hash
            return True
        except xmlrpc.client.Fault as err:
            if err.faultCode != _XMLRPC_UNKNOWN_HASH:
                raise
        return False

    def get_hashes(self) -> List[str]:
        return list(map(str.lower, self.__server.download_list()))

    @hash_or_torrent
    @_catch_unknown_torrent
    def get_torrent_path(self, torrent_hash: str) -> str:
        return self.__server.d.get_loaded_file(torrent_hash)

    @hash_or_torrent
    @_catch_unknown_torrent
    def get_data_prefix(self, torrent_hash: str) -> str:
        mc = xmlrpc.client.MultiCall(self.__server)
        mc.d.get_directory(torrent_hash)
        mc.d.is_multi_file(torrent_hash)
        (path, is_multi_file) = mc()
        if is_multi_file:
            path = os.path.dirname(os.path.normpath(path))
        return path

    def get_data_prefix_default(self) -> str:
        return self.__server.get_directory()

    # ===

    @hash_or_torrent
    @_catch_unknown_torrent
    def get_full_path(self, torrent_hash: str) -> str:
        mc = xmlrpc.client.MultiCall(self.__server)
        mc.d.get_directory(torrent_hash)
        mc.d.get_name(torrent_hash)
        mc.d.is_multi_file(torrent_hash)
        (path, name, is_multi_file) = mc()
        if is_multi_file:
            return path
        return os.path.join(path, name)

    @hash_or_torrent
    @_catch_unknown_torrent
    def get_file_name(self, torrent_hash: str) -> str:
        return self.__server.d.get_name(torrent_hash)

    @hash_or_torrent
    @_catch_unknown_torrent
    def is_single_file(self, torrent_hash: str) -> bool:
        return (not self.__server.d.is_multi_file(torrent_hash))

    @hash_or_torrent
    @_catch_unknown_torrent
    def get_files(self, torrent_hash: str, on_fs: bool=False) -> Dict[str, TorrentEntryAttrs]:
        mc = xmlrpc.client.MultiCall(self.__server)
        mc.d.get_base_path(torrent_hash)
        mc.d.get_base_filename(torrent_hash)
        mc.d.is_multi_file(torrent_hash)
        mc.d.get_size_files(torrent_hash)
        mc.f.get_size_bytes(torrent_hash, 0)
        (base_path, base_file_name, is_multi_file, count, first_file_size) = tuple(mc())
        base = (base_path if on_fs else base_file_name)

        if not is_multi_file:
            return {base: TorrentEntryAttrs.new_file(first_file_size)}

        mc = xmlrpc.client.MultiCall(self.__server)
        for index in range(count):
            mc.f.get_path(torrent_hash, index)
            mc.f.get_size_bytes(torrent_hash, index)
        flist = list(mc())
        flist = list(zip(flist[::2], flist[1::2]))

        files = build_files(base, flist)
        files.update({base: TorrentEntryAttrs.new_dir()})
        return files

    # ===

    @classmethod
    def get_custom_keys(cls) -> List[str]:
        return ["c1", "c2", "c3", "c4", "c5"]

    @hash_or_torrent
    @_catch_unknown_torrent
    def set_customs(self, torrent_hash: str, customs: Dict[str, str]) -> None:
        assert len(customs) != 0, "Empty customs dict"
        mc = xmlrpc.client.MultiCall(self.__server)
        for (key, value) in customs.items():
            getattr(mc.d, "set_custom{}".format(key[1:]))(torrent_hash, value)
        mc()

    @hash_or_torrent
    @_catch_unknown_torrent
    def get_customs(self, torrent_hash: str, keys: List[str]) -> Dict[str, str]:
        assert len(keys) != 0, "Empty customs keys list"
        keys = list(set(keys))
        mc = xmlrpc.client.MultiCall(self.__server)
        for key in keys:
            getattr(mc.d, "get_custom{}".format(key[1:]))(torrent_hash)
        return dict(zip(keys, list(mc())))



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import socket
import smtplib
import email.mime.multipart
import email.mime.text
import email.header
import email.utils
import contextlib
import time

from typing import List
from typing import Dict
from typing import Union
from typing import Any

from ...optconf import Option
from ...optconf.converters import as_string_list
from ...optconf.converters import as_path_or_empty

from . import ResultsType
from . import BaseConfetti
from . import templated


# =====
class Plugin(BaseConfetti):  # pylint: disable=too-many-instance-attributes
    PLUGIN_NAME = "email"

    def __init__(  # pylint: disable=super-init-not-called,too-many-arguments,too-many-locals
        self,
        to: str,
        cc: List[str],
        subject: str,
        sender: str,
        html: bool,
        template: str,
        server: str,
        port: int,
        ssl: bool,
        user: str,
        passwd: str,
        timeout: float,
        retries: int,
        retries_sleep: float,
        **kwargs: Any,
    ) -> None:

        self._init_bases(**kwargs)

        self.__to = to
        self.__cc = cc
        self.__subject = subject
        self.__sender = sender
        self.__html = html
        self.__template_path = template

        self.__server = server
        self.__port = port
        self.__ssl = ssl
        self.__user = user
        self.__passwd = passwd

        self.__timeout = timeout
        self.__retries = retries
        self.__retries_sleep = retries_sleep

    @classmethod
    def get_options(cls) -> Dict[str, Option]:
        return cls._get_merged_options({
            "to":       Option(default=["root@localhost"], type=as_string_list, help="Destination email address"),
            "cc":       Option(default=[], type=as_string_list, help="Email 'CC' field"),
            "subject":  Option(default="{source} report: you have {affected} new torrents ^_^", help="Email subject"),
            "sender":   Option(default="root@localhost", help="Email 'From' field"),
            "html":     Option(default=True, help="HTML or plaintext email body"),
            "template": Option(default="", type=as_path_or_empty, help="Mako template file name"),

            "server":   Option(default="localhost", help="Hostname of SMTP server"),
            "port":     Option(default=0, help="Port of SMTP server"),
            "ssl":      Option(default=False, help="Use SMTPS"),
            "user":     Option(default="", help="Account on SMTP server"),
            "passwd":   Option(default="", help="Passwd for account on SMTP server"),

            "timeout":       Option(default=10.0, help="Network timeout"),
            "retries":       Option(default=5, help="Retries for failed attempts"),
            "retries_sleep": Option(default=1.0, help="Sleep interval between failed attempts"),
        })

    # ===

    def send_results(self, source: str, results: ResultsType) -> None:
        msg = self.__format_message(source, results)
        retries = self.__retries
        while True:
            try:
                self.__send_message(msg)
                break
            except (
                smtplib.SMTPServerDisconnected,
                smtplib.SMTPConnectError,
                smtplib.SMTPHeloError,
                socket.timeout,
            ):
                if retries == 0:
                    raise
                time.sleep(self.__retries_sleep)
                retries -= 1

    # ===

    def __format_message(self, source: str, results: ResultsType) -> email.mime.multipart.MIMEMultipart:
        subject_placeholders: Dict[str, Union[str, int]] = {
            field: len(items)
            for (field, items) in results.items()
        }
        subject_placeholders["source"] = source
        return self.__make_message(
            subject=self.__subject.format(**subject_placeholders),
            body=templated(
                name=(self.__template_path if self.__template_path else "email.{ctype}.{source}.mako").format(
                    ctype=("html" if self.__html else "plain"),
                    source=source,
                ),
                built_in=(not self.__template_path),
                source=source,
                results=results,
            ),
        )

    def __make_message(self, subject: str, body: str) -> email.mime.multipart.MIMEMultipart:
        email_headers = {
            "From":    self.__sender,
            "To":      ", ".join(self.__to),
            "Date":    email.utils.formatdate(localtime=True),
            "Subject": email.header.Header(subject, "utf-8"),
        }
        if len(self.__cc) > 0:
            email_headers["CC"] = ", ".join(self.__cc)

        msg = email.mime.multipart.MIMEMultipart()
        for (key, value) in email_headers.items():
            msg[key] = value  # type: ignore

        msg.attach(email.mime.text.MIMEText(  # type: ignore
            _text=body.encode("utf-8"),
            _subtype=("html" if self.__html else "plain"),
            _charset="utf-8",
        ))
        return msg

    def __send_message(self, msg: email.mime.multipart.MIMEMultipart) -> None:
        smtp_class = (smtplib.SMTP_SSL if self.__ssl else smtplib.SMTP)
        with contextlib.closing(smtp_class(
            host=self.__server,
            port=self.__port,
            timeout=self.__timeout,
        )) as client:
            if self.__user:
                client.login(self.__user, self.__passwd)  # pylint: disable=no-member
            client.send_message(msg)  # pylint: disable=no-member



import os

from typing import List
from typing import Dict
from typing import Sequence
from typing import Union
from typing import Any


# =====
def as_string_list(values: Union[str, Sequence]) -> List[str]:
    if isinstance(values, str):
        values = [values]
    return list(map(str, values))


def as_key_value(values: Union[str, Dict[str, Any]]) -> Dict[str, str]:
    if isinstance(values, dict):
        return values
    return dict(
        tuple(map(str.strip, (item.split("=", 1) + [""])[:2]))  # type: ignore
        for item in as_string_list(values)
        if len(item.split("=", 1)[0].strip()) != 0
    )


def as_path(value: str) -> str:
    return os.path.normpath(os.path.abspath(os.path.expanduser(value)))


def as_paths_list(values: Sequence[str]) -> List[str]:
    if isinstance(values, str):
        values = [values]
    return list(map(as_path, values))


def as_path_or_empty(value: str) -> str:
    if value:
        return as_path(value)
    return ""


def as_8int(value: int) -> int:
    return int(str(value), 8)



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import json
import traceback
import operator
import argparse

from typing import List
from typing import Dict
from typing import NamedTuple
from typing import Optional

from ..plugins.trackers import TrackerError
from ..plugins.trackers import TrackerStat
from ..plugins.trackers import WithStat

from ..helpers import tcollection

from ..tfile import Torrent

from ..cli import CellAlign
from ..cli import Cell
from ..cli import Log

from . import init
from . import wrap_main
from . import get_configured_log
from . import get_configured_trackers


# =====
class _InnerStatRecord(NamedTuple):
    name: str
    status: str
    torrent: Optional[Torrent]
    tracker: Optional[WithStat]
    stat: TrackerStat
    err_name: str
    err_msg: str
    tb_lines: List[str]


class StatRecord(_InnerStatRecord):
    @staticmethod
    def new(
        name: str,
        status: str,
        torrent: Optional[Torrent] = None,
        tracker: Optional[WithStat] = None,
        stat: Optional[TrackerStat] = None,
        err_name: str="",
        err_msg: str="",
        tb_lines: Optional[List[str]] = None,
    ) -> "StatRecord":

        return StatRecord(
            name=name,
            status=status,
            torrent=torrent,
            tracker=tracker,
            stat=(stat or TrackerStat()),
            err_name=err_name,
            err_msg=err_msg,
            tb_lines=(tb_lines or []),
        )


def select_tracker(trackers: List[WithStat], torrent: Torrent) -> Optional[WithStat]:
    for tracker in trackers:
        if tracker.is_matched_for(torrent):
            return tracker  # type: ignore
    return None


def fetch_stat(
    trackers: List[WithStat],
    torrents: Dict[str, Optional[Torrent]],
    log: Log,
) -> List[StatRecord]:

    if not log.isatty():
        log.info("Fetching statistics ...")

    stats: List[StatRecord] = []
    for (name, torrent) in log.progress(
        sorted(torrents.items(), key=operator.itemgetter(0)),
        ("Fetching statistics", ()),
        ("Fetched statistics for {magenta}%d{reset} torrents", (lambda: len(stats),)),
    ):
        if torrent:
            tracker = select_tracker(trackers, torrent)
            if tracker:
                try:
                    record = StatRecord.new(
                        name=name,
                        status="passed",
                        torrent=torrent,
                        tracker=tracker,
                        stat=tracker.fetch_stat(torrent)
                    )
                except Exception as err:
                    record = StatRecord.new(
                        name=name,
                        status=("tracker_error" if isinstance(err, TrackerError) else "unhandled_error"),
                        torrent=torrent,
                        tracker=tracker,
                        err_name=type(err).__name__,
                        err_msg=str(err),
                        tb_lines="".join(traceback.format_exc()).strip().split("\n"),
                    )
            else:
                record = StatRecord.new(name=name, status="unknown", torrent=torrent)
        else:
            record = StatRecord.new(name=name, status="invalid")
        stats.append(record)

    if not log.isatty():
        log.info("Fetched statistics for {magenta}%d{reset} torrents", (len(stats),))
    return stats


def print_stats_table(stats: List[StatRecord], min_seeders: int, log: Log) -> None:
    log.print()
    log.print_table(
        header=[
            Cell(data=col, align=CellAlign.CENTER)
            for col in ["Name", "Tracker", "Seeders", "Leechers", "Status", "Error"]
        ],
        table=[
            [
                Cell(record.name),
                Cell(record.tracker.PLUGIN_NAME if record.tracker else ""),
                Cell(
                    data=str(record.stat.seeders),
                    colors=("{green}" if record.stat.seeders >= min_seeders else "{red}"),
                    align=CellAlign.RIGHT,
                ),
                Cell(
                    data=str(record.stat.leechers),
                    colors=("{yellow}" if record.stat.leechers else ""),
                    align=CellAlign.RIGHT,
                ),
                Cell(
                    data=record.status,
                    colors=("{green}" if record.status == "passed" else "{red}"),
                ),
                Cell(record.err_msg if record.status == "tracker_error" else ""),
            ]
            for record in sorted(stats, key=(lambda record: (-record.stat.seeders, record.name)))
        ],
    )
    log.print()


def export_stats(stats: List[StatRecord], path: str) -> None:
    with open(path, "w") as export_file:
        json.dump({
            record.name: {
                "status": record.status,
                "torrent": ({
                    "hash": record.torrent.get_hash(),
                    "path": record.torrent.get_path(),
                    "comment": record.torrent.get_comment(),
                } if record.torrent else None),
                "tracker": (record.tracker.PLUGIN_NAME if record.tracker else ""),
                "stat": record.stat._asdict(),
                "err_name": record.err_name,
                "err_msg": record.err_msg,
                "tb_lines": record.tb_lines,
            }
            for record in stats
        }, export_file, indent=4, sort_keys=True)


# ===== Main =====
@wrap_main
def main() -> None:
    (parent_parser, argv, config) = init()
    args_parser = argparse.ArgumentParser(
        prog="emstat",
        description="Collect torrents statistics from trackers",
        parents=[parent_parser],
    )
    args_parser.add_argument("-f", "--name-filter", default="", metavar="<wildcard_pattern>")
    args_parser.add_argument("-y", "--only-trackers", default=[], nargs="+", metavar="<tracker>")
    args_parser.add_argument("-x", "--exclude-trackers", default=[], nargs="+", metavar="<tracker>")
    args_parser.add_argument("--fail-on-captcha", action="store_true")
    args_parser.add_argument("--export", default="", metavar="<json_file>")
    options = args_parser.parse_args(argv[1:])

    with get_configured_log(config, False, sys.stdout) as log_stdout:
        with get_configured_log(config, False, sys.stderr) as log_stderr:

            def read_captcha(url: str) -> str:
                if options.fail_on_captcha:
                    raise RuntimeError("Required decoding of captcha but '--fail-on-captcha' specified")
                else:
                    log_stderr.info("{yellow}Enter the captcha{reset} from [{blue}%s{reset}]: ", (url,), no_nl=True)
                    return input()

            trackers: List[WithStat] = get_configured_trackers(  # type: ignore
                config=config,
                captcha_decoder=read_captcha,
                only=options.only_trackers,
                exclude=options.exclude_trackers,
                required_bases=[WithStat],
                fail_bad_tracker=config.emstat.fail_bad_tracker,
                log=log_stderr,
            )

            torrents = tcollection.load_from_dir(
                path=config.core.torrents_dir,
                name_filter=(options.name_filter or config.emupdate.name_filter),
                precalculate_hashes=bool(options.export),
                log=log_stderr,
            )

            stats = fetch_stat(
                trackers=trackers,
                torrents=torrents,
                log=log_stderr,
            )

            print_stats_table(stats, config.emstat.min_seeders, log_stdout)
            if options.export:
                export_stats(stats, options.export)


if __name__ == "__main__":
    main()  # Do the thing!



"""
    Emonoda -- A set of tools to organize and manage your torrents
    Copyright (C) 2015  Devaev Maxim <mdevaev@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import sys
import os
import contextlib
import argparse

from typing import TextIO
from typing import Tuple
from typing import List
from typing import Dict
from typing import Sequence
from typing import Callable
from typing import Generator
from typing import Optional
from typing import Type
from typing import Any

import pygments
import pygments.lexers.data
import pygments.formatters

from ..optconf import make_config
from ..optconf import Section
from ..optconf import Option

from ..optconf import build_raw_from_options
from ..optconf.dumper import make_config_dump
from ..optconf.loader import load_file as load_yaml_file
from ..optconf.converters import as_string_list
from ..optconf.converters import as_key_value
from ..optconf.converters import as_path
from ..optconf.converters import as_paths_list
from ..optconf.converters import as_path_or_empty
from ..optconf.converters import as_8int

from ..plugins.clients import get_client_class
from ..plugins.trackers import get_tracker_class
from ..plugins.confetti import get_confetti_class

from ..plugins.trackers import BaseTracker
from ..plugins.trackers import WithLogin as T_WithLogin
from ..plugins.trackers import WithCaptcha as T_WithCaptcha
from ..plugins.trackers import WithCheckTime as T_WithCheckTime
from ..plugins.clients import BaseClient
from ..plugins.clients import WithCustoms as C_WithCustoms
from ..plugins.confetti import BaseConfetti

from ..cli import Log


# =====
class StoreTrueOrderedAction(argparse.Action):
    def __init__(
        self,
        option_strings: Sequence[str],
        dest: str,
        default: bool=False,
        required: bool=False,
        help: Optional[str]=None,  # pylint: disable=redefined-builtin
    ) -> None:

        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=0,
            const=True,
            default=default,
            required=required,
            help=help,
        )

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: Any=None,
    ) -> None:

        if not hasattr(namespace, "ordered_flags"):
            setattr(namespace, "ordered_flags", [])
        namespace.ordered_flags.append((self.dest, self.const))


def init() -> Tuple[argparse.ArgumentParser, List[str], Section]:
    args_parser = argparse.ArgumentParser(add_help=False)
    args_parser.add_argument("-c", "--config", dest="config_file_path", default="~/.config/emonoda.yaml", metavar="<file>")
    args_parser.add_argument("-o", "--set-options", dest="set_options", default=[], nargs="+")
    args_parser.add_argument("-m", "--dump-config", dest="dump_config", action="store_true")
    (options, remaining) = args_parser.parse_known_args(sys.argv)

    options.config_file_path = os.path.expanduser(options.config_file_path)
    if os.path.exists(options.config_file_path):
        raw_config = load_yaml_file(options.config_file_path)
    else:
        raw_config = {}
    _merge_dicts(raw_config, build_raw_from_options(options.set_options))
    scheme = _get_config_scheme()
    config = make_config(raw_config, scheme)

    if config.core.client:
        client_scheme = get_client_class(config.core.client).get_options()
        scheme["client"] = client_scheme

    for tracker_name in raw_config.get("trackers", []):
        tracker_scheme = get_tracker_class(tracker_name).get_options()
        scheme.setdefault("trackers", {})
        scheme["trackers"][tracker_name] = tracker_scheme

    for confetti_name in raw_config.get("confetti", []):
        confetti_scheme = get_confetti_class(confetti_name).get_options()
        scheme.setdefault("confetti", {})
        scheme["confetti"][confetti_name] = confetti_scheme

    config = make_config(raw_config, scheme)

    if options.dump_config:
        dump = make_config_dump(config)
        if sys.stdout.isatty():
            dump = pygments.highlight(
                dump,
                pygments.lexers.data.YamlLexer(),
                pygments.formatters.TerminalFormatter(bg="dark"),
            )
        print(dump)
        sys.exit(0)

    config.setdefault("client", Section())
    config.setdefault("trackers", Section())
    config.setdefault("confetti", Section())

    return (args_parser, remaining, config)


def wrap_main(method: Callable[..., None]) -> Callable[..., None]:
    def wrap() -> None:
        try:
            method()
        except (SystemExit, KeyboardInterrupt):
            sys.exit(1)
    return wrap


def _merge_dicts(dest: Dict, src: Dict) -> None:
    for key in src:
        if key in dest:
            if isinstance(dest[key], dict) and isinstance(src[key], dict):
                _merge_dicts(dest[key], src[key])
                continue
        dest[key] = src[key]


# =====
@contextlib.contextmanager
def get_configured_log(config: Section, quiet: bool, output: TextIO) -> Generator[Log, None, None]:
    log = Log(config.core.use_colors, config.core.force_colors, quiet, output)
    try:
        yield log
    finally:
        log.finish()


def get_configured_client(config: Section, required: bool, with_customs: bool, log: Log) -> Optional[BaseClient]:
    name = config.core.client
    if name:
        log.info("Enabling the client {blue}%s{reset} ...", (name,), one_line=True)
        try:
            cls = get_client_class(name)
            if with_customs and C_WithCustoms not in cls.get_bases():
                raise RuntimeError("Your client does not support custom fields")
            client = cls(**config.client)
        except Exception as err:
            log.error("Can't init client {red}%s{reset}: {red}%s{reset}(%s)", (name, type(err).__name__, err))
            raise
        log.info("Client {blue}%s{reset} is {green}ready{reset}", (name,))
        return client
    elif required:
        raise RuntimeError("No configured client found")
    else:
        return None


def get_configured_trackers(
    config: Section,
    captcha_decoder: Callable[[str], str],
    only: List[str],
    exclude: List[str],
    required_bases: List[Type[BaseTracker]],
    fail_bad_tracker: bool,
    log: Log,
) -> List[BaseTracker]:

    to_init = set(config.trackers).difference(exclude)
    if len(only) != 0:
        to_init = to_init.intersection(only)

    for name in list(to_init):
        if not set(required_bases).issubset(get_tracker_class(name).get_bases()):
            to_init.remove(name)

    if len(to_init) == 0:
        raise RuntimeError("No trackers to init")

    trackers = []
    for name in sorted(to_init):
        log.info("Enabling the tracker {blue}%s{reset} ...", (name,), one_line=True)

        cls = get_tracker_class(name)
        bases = cls.get_bases()
        kwargs = dict(config.trackers[name])
        if T_WithCaptcha in cls.get_bases():
            kwargs["captcha_decoder"] = captcha_decoder
        tracker = cls(**kwargs)

        try:
            log.info("Enabling the tracker {blue}%s{reset}: {yellow}testing{reset} ...", (name,), one_line=True)
            tracker.test()
            if T_WithLogin in bases:
                log.info("Enabling the tracker {blue}%s{reset}: {yellow}logging in{reset} ...", (name,), one_line=True)
                tracker.login()  # type: ignore
            if T_WithCheckTime in bases:
                log.info("Enabling the tracker {blue}%s{reset}: {yellow}configuring timezone{reset} ...", (name,), one_line=True)
                tracker.init_tzinfo()  # type: ignore
            log.info("Tracker {blue}%s{reset} is {green}ready{reset}", (name,))
        except Exception as err:
            log.error("Can't init tracker {red}%s{reset}: {red}%s{reset}(%s)", (name, type(err).__name__, err))
            if fail_bad_tracker:
                raise
            continue

        trackers.append(tracker)

    return trackers


def get_configured_confetti(config: Section, log: Log) -> List[BaseConfetti]:
    senders: List[BaseConfetti] = []
    for name in sorted(config.confetti):
        log.info("Enabling the confetti {blue}%s{reset} ...", (name,), one_line=True)
        cls = get_confetti_class(name)
        try:
            senders.append(cls(**config.confetti[name]))
            log.info("Confetti {blue}%s{reset} is {green}ready{reset}", (name,))
        except Exception as err:
            log.error("Can't init confetti {red}%s{reset}: {red}%s{reset}(%s)", (name, type(err).__name__, err))
            raise
    return senders


# =====
def _get_config_scheme() -> Dict:
    return {
        "core": {
            "client":        Option(default="", help="The name of plugin for torrent client"),
            "torrents_dir":  Option(default=".", type=as_path, help="Path to directory with torrent files"),
            "data_root_dir": Option(default="~/Downloads", type=as_path_or_empty, help="Path to root directory with data of torrents"),
            "another_data_root_dirs": Option(default=[], type=as_paths_list, help="Paths to another data directories"),
            "use_colors":    Option(default=True, help="Enable colored output"),
            "force_colors":  Option(default=False, help="Always use the coloring"),
        },

        "emupdate": {
            "name_filter":   Option(default="*.torrent", help="Update only filtered torrent files"),
            "backup_dir":    Option(default="", type=as_path_or_empty, help="Backup old torrent files after update here"),
            "backup_suffix": Option(default=".%Y.%m.%d-%H:%M:%S.bak", help="Append this suffix to backuped file"),
            "save_customs":  Option(default=[], type=as_string_list, help="Save client custom fields after update if supports"),
            "set_customs":   Option(default={}, type=as_key_value, help="Set client custom fileds after update if supports"),
            "show_unknown":  Option(default=False, help="Show the torrents with unknown tracker in the log"),
            "show_passed":   Option(default=False, help="Show the torrents without changes"),
            "show_diff":     Option(default=True, help="Show diff between old and updated torrent files"),
            "fail_bad_tracker": Option(default=True, help="Fail on trackers with invalid configuration"),
        },

        "emstat": {
            "min_seeders":      Option(default=5, help="Minimal seeders to green color"),
            "fail_bad_tracker": Option(default=True, help="Fail on trackers with invalid configuration"),
        },

        "emfile": {
            "show_customs": Option(default=[], type=as_string_list, help="Show custom fields from client"),
        },

        "emload": {
            "mkdir_mode":  Option(default=-1, type=as_8int, help="Permission for new directories"),
            "set_customs": Option(default={}, type=as_key_value, help="Set client custom fileds after update if supports")
        },

        "emfind": {
            "cache_file":  Option(default="~/.cache/emfind.pk", type=as_path, help="Torrents cache"),
            "name_filter": Option(default="*.torrent", help="Cache only filtered torrent files"),
            "ignore_orphans": Option(default=[], type=as_paths_list, help="Ignore these paths on the final analyse"),
        },
    }
